@@ -5,6 +5,7 @@ import warnings
 
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidExpressionError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
+from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidObjectError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLMissingCPFDefinitionError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLRepeatedVariableError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
@@ -14,8 +15,6 @@ from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLValueOutOfRangeError
 from pyRDDLGym.Core.Compiler.RDDLModel import RDDLGroundedModel
 from pyRDDLGym.Core.Parser.expr import Expression
 
-PRIME = '\''
-
 AGGREG_OP_TO_STRING_DICT = {
     'prod': '*',
     'sum': '+',
@@ -67,8 +66,7 @@ class RDDLGrounder(Grounder):
         self.invariants = []
         
         self.index_of_object = {}
-        self.enum_types = set()
-        self.enum_literals = set()
+        self.enums = set()
         
         self.param_types = {}
         self.variable_types = {}
@@ -123,8 +121,7 @@ class RDDLGrounder(Grounder):
         model.variable_ranges = self.variable_ranges
         
         model.index_of_object = self.index_of_object
-        model.enum_types = self.enum_types
-        model.enum_literals = self.enum_literals        
+        model.enums = self.enums   
         model.grounded_names = self.grounded_names
         
         return model
@@ -160,8 +157,7 @@ class RDDLGrounder(Grounder):
                 for obj in obj_type[1]:
                     self.objects_rev[obj] = obj_type[0]
         self.index_of_object = {}
-        self.enum_types = set()
-        self.enum_literals = set()
+        self.enums = set()
 
     def _ground_objects(self, args):
         objects_by_type = []
@@ -174,11 +170,12 @@ class RDDLGrounder(Grounder):
         return itertools.product(*objects_by_type)
     
     def _append_variation_to_name(self, base_name, variation):
-        return base_name + '_' + '_'.join(variation)
+        objects = RDDLGroundedModel.OBJECT_SEP.join(variation)
+        return base_name + RDDLGroundedModel.FLUENT_SEP + objects
         
-    def _generate_grounded_names(self, base_name,
-                                 variation_list,
+    def _generate_grounded_names(self, base_name, variation_list,
                                  return_grounding_param_dict=False):
+        PRIME = RDDLGroundedModel.NEXT_STATE_SYM
         all_grounded_names = []
         prime_var = PRIME in base_name
         base_name = base_name.strip(PRIME)
@@ -238,6 +235,7 @@ class RDDLGrounder(Grounder):
                 self.cpforder[level] = [pvariable.name]
 
     def _ground_pvariables_and_cpf(self):
+        PRIME = RDDLGroundedModel.NEXT_STATE_SYM
         all_grounded_state_cpfs = []
         all_grounded_interim_cpfs = []
         all_grounded_derived_cpfs = []
@@ -591,8 +589,18 @@ class RDDLGrounder(Grounder):
                     
     # added on Dec 17 (Mike)
     def _extract_variable_information(self):
+        PRIME = RDDLGroundedModel.NEXT_STATE_SYM
         var_params, var_types, var_ranges = {}, {}, {}
         for pvar in self.AST.domain.pvariables:
+            
+            # make sure name does not contain separators
+            SEPARATORS = [RDDLGroundedModel.FLUENT_SEP, RDDLGroundedModel.OBJECT_SEP] 
+            for separator in SEPARATORS:
+                if separator in pvar.name:
+                    raise RDDLInvalidObjectError(
+                        f'Variable name <{pvar.name}> contains the '
+                        f'illegal separator {separator}.')
+                    
             objects = pvar.param_types
             if objects is None:
                 objects = []
@@ -605,7 +613,7 @@ class RDDLGrounder(Grounder):
                 if name in var_params:
                     raise RDDLRepeatedVariableError(
                         f'{pvar.fluent_type} <{pvar.name}> has the same name as '
-                        f'another {var_types[name]}.')
+                        f'another pvariable {var_types[name]}.')
                 primed_name = (name + PRIME) if pvar.is_state_fluent() else name
                 var_params[name] = var_params[primed_name] = []        
                 var_types[name] = pvar.fluent_type
