@@ -1,20 +1,20 @@
-import numpy as np
+from functools import partial
 import jax
 import jax.numpy as jnp
 import jax.random as random
 import jax.scipy as scipy 
-from typing import Dict
+from tensorflow_probability.substrates import jax as tfp
+from typing import Callable, Dict
 
+from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError
 
-from pyRDDLGym.Core.Compiler.RDDLDecompiler import RDDLDecompiler
 from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
 from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
 from pyRDDLGym.Core.Compiler.RDDLObjectsTracer import RDDLObjectsTracer
 from pyRDDLGym.Core.Compiler.RDDLValueInitializer import RDDLValueInitializer
 from pyRDDLGym.Core.Debug.Logger import Logger
-from pyRDDLGym.Core.Parser.expr import Expression
 
 
 class JaxRDDLCompiler:
@@ -43,7 +43,7 @@ class JaxRDDLCompiler:
         '''
         self.rddl = rddl
         self.logger = logger
-        jax.config.update('jax_log_compiles', True)
+        # jax.config.update('jax_log_compiles', True) # for testing ONLY
         
         # compile initial values
         if self.logger is not None:
@@ -121,11 +121,8 @@ class JaxRDDLCompiler:
             'min': jnp.minimum,
             'max': jnp.maximum,
             'pow': jnp.power,
-            'log': lambda x, y: jnp.log(x) / jnp.log(y)
-        }        
-        self.CONTROL_OPS = {
-            'if': jnp.where,
-            'switch': self._jax_switch_helper()
+            'log': lambda x, y: jnp.log(x) / jnp.log(y),
+            'hypot': jnp.hypot
         }
         
     # ===========================================================================
@@ -180,16 +177,16 @@ class JaxRDDLCompiler:
     def _compile_reward(self):
         return self._jax(self.rddl.reward, dtype=JaxRDDLCompiler.REAL)
     
-    def compile_rollouts(self, policy, n_steps: int, n_batch: int,
+    def compile_rollouts(self, policy: Callable, 
+                         n_steps: int, n_batch: int,
                          check_constraints: bool=False):
         '''Compiles the current RDDL into a wrapped function that samples multiple
         rollouts (state trajectories) in batched form for the given policy. The
         wrapped function takes the policy parameters and RNG key as input, and
         returns a dictionary of all logged information from the rollouts.
         
-        :param policy: a Jax compiled function that takes the subs dict, step
-        number, policy parameters and an RNG key and returns an action and the
-        next RNG key in the sequence
+        :param policy: a Jax compiled function that takes the policy parameters, 
+        decision epoch, state dict, and an RNG key and returns an action dict
         :param n_steps: the length of each rollout
         :param n_batch: how many rollouts each batch performs
         :param check_constraints: whether state, action and termination 
@@ -197,84 +194,92 @@ class JaxRDDLCompiler:
         returned log and does not raise an exception
         '''
         NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
-            
-        # compute a batched version of the initial values
-        init_subs = {}
-        for (name, value) in self.init_values.items():
-            new_shape = (n_batch,) + np.shape(value)
-            init_subs[name] = np.broadcast_to(value, shape=new_shape)            
-        for (state, next_state) in self.rddl.next_state.items():
-            init_subs[next_state] = init_subs[state]
         
-        # this performs a single step update
-        def _step(carried, step):
-            subs, params, key = carried
-            action, key = policy(subs, step, params, key)
-            subs.update(action)
-            error = NORMAL
+        rddl = self.rddl
+        reward_fn, cpfs = self.reward, self.cpfs
+        preconds, invariants, terminals = \
+            self.preconditions, self.invariants, self.termination
+        
+        # do a single step update from the RDDL model
+        def _jax_wrapped_single_step(key, params, step, subs):
+            errors = NORMAL
             
-            # check preconditions
+            # compute action
+            key, subkey = random.split(key)
+            states = {var: values 
+                      for (var, values) in subs.items()
+                      if rddl.variable_types[var] == 'state-fluent'}
+            actions = policy(subkey, params, step, states)
+            subs.update(actions)
+            
+            # check action preconditions
+            precond_check = True
             if check_constraints:
-                preconds = jnp.zeros(shape=(len(self.preconditions),), dtype=bool)
-                for (i, precond) in enumerate(self.preconditions):
-                    sample, key, precond_err = precond(subs, key)
-                    preconds = preconds.at[i].set(sample)
-                    error |= precond_err
+                for precond in preconds:
+                    sample, key, err = precond(subs, key)
+                    precond_check = jnp.logical_and(precond_check, sample)
+                    errors |= err
             
-            # calculate CPFs in topological order and reward
-            for (name, cpf) in self.cpfs.items():
-                subs[name], key, cpf_err = cpf(subs, key)
-                error |= cpf_err                
-            reward, key, reward_err = self.reward(subs, key)
-            error |= reward_err
+            # calculate CPFs in topological order
+            for (name, cpf) in cpfs.items():
+                subs[name], key, err = cpf(subs, key)
+                errors |= err                
+                
+            # calculate the immediate reward
+            reward, key, err = reward_fn(subs, key)
+            errors |= err
             
-            for (state, next_state) in self.rddl.next_state.items():
+            # set the next state to the current state
+            for (state, next_state) in rddl.next_state.items():
                 subs[state] = subs[next_state]
             
-            # check the invariants in the new state
+            # check the state invariants
+            invariant_check = True
             if check_constraints:
-                invariants = jnp.zeros(shape=(len(self.invariants),), dtype=bool)
-                for (i, invariant) in enumerate(self.invariants):
-                    sample, key, invariant_err = invariant(subs, key)
-                    invariants = invariants.at[i].set(sample)
-                    error |= invariant_err
+                for invariant in invariants:
+                    sample, key, err = invariant(subs, key)
+                    invariant_check = jnp.logical_and(invariant_check, sample)
+                    errors |= err
             
             # check the termination (TODO: zero out reward in s if terminated)
+            terminated_check = False
             if check_constraints:
-                terminated = False
-                for terminal in self.termination:
-                    sample, key, terminal_err = terminal(subs, key)
-                    terminated = jnp.logical_or(terminated, sample)
-                    error |= terminal_err
+                for terminal in terminals:
+                    sample, key, err = terminal(subs, key)
+                    terminated_check = jnp.logical_or(terminated_check, sample)
+                    errors |= err
             
-            logged = {'fluent': subs,
-                      'action': action,
-                      'reward': reward,
-                      'error': error}            
-            if check_constraints:
-                logged['preconditions'] = preconds
-                logged['invariants'] = invariants
-                logged['terminated'] = terminated
-                
-            carried = (subs, params, key)
-            return carried, logged
+            log = {
+                'pvar': subs,
+                'action': actions,
+                'reward': reward,
+                'error': errors,
+                'precondition': precond_check,
+                'invariant': invariant_check,
+                'termination': terminated_check
+            }
+            return log, subs
         
-        # this performs a single roll-out starting from subs
-        def _rollout(subs, params, key):
+        # do a batched step update from the RDDL model
+        def _jax_wrapped_batched_step(carry, step):
+            key, params, subs = carry  
+            key, *subkeys = random.split(key, num=1 + n_batch)
+            batched_step = jax.vmap(
+                _jax_wrapped_single_step, in_axes=(0, None, None, 0)
+            ) 
+            log, subs = batched_step(jnp.asarray(subkeys), params, step, subs)            
+            carry = (key, params, subs)
+            return carry, log            
+            
+        # do a batched roll-out from the RDDL model
+        def _jax_wrapped_batched_rollout(key, params, subs):
+            start = (key, params, subs)
             steps = jnp.arange(n_steps)
-            carry = (subs, params, key)
-            (* _, key), logged = jax.lax.scan(_step, carry, steps)
-            return logged, key
+            _, log = jax.lax.scan(_jax_wrapped_batched_step, start, steps)
+            log = jax.tree_map(partial(jnp.swapaxes, axis1=0, axis2=1), log)
+            return log
         
-        # this performs batched roll-outs
-        def _rollouts(params, key): 
-            subkeys = jax.random.split(key, num=n_batch)
-            logged, keys = jax.vmap(_rollout, in_axes=(0, None, 0))(
-                init_subs, params, subkeys)
-            logged['keys'] = subkeys
-            return logged, keys
-        
-        return _rollouts
+        return _jax_wrapped_batched_rollout
     
     # ===========================================================================
     # error checks
@@ -298,14 +303,6 @@ class JaxRDDLCompiler:
                                    for expr in self.termination]
         return printed
         
-    @staticmethod
-    def _print_stack_trace(expr):
-        if isinstance(expr, Expression):
-            trace = RDDLDecompiler().decompile_expr(expr)
-        else:
-            trace = str(expr)
-        return '>> ' + trace
-    
     @staticmethod
     def _check_valid_op(expr, valid_ops):
         etype, op = expr.etype
@@ -313,8 +310,7 @@ class JaxRDDLCompiler:
             valid_op_str = ','.join(valid_ops.keys())
             raise RDDLNotImplementedError(
                 f'{etype} operator {op} is not supported: '
-                f'must be in {valid_op_str}.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))
+                f'must be in {valid_op_str}.\n' + print_stack_trace(expr))
     
     @staticmethod
     def _check_num_args(expr, required_args):
@@ -323,8 +319,7 @@ class JaxRDDLCompiler:
             etype, op = expr.etype
             raise RDDLInvalidNumberOfArgumentsError(
                 f'{etype} operator {op} requires {required_args} arguments, '
-                f'got {actual_args}.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))
+                f'got {actual_args}.\n' + print_stack_trace(expr))
         
     ERROR_CODES = {
         'NORMAL': 0,
@@ -345,11 +340,14 @@ class JaxRDDLCompiler:
         'INVALID_PARAM_CAUCHY': 16384,
         'INVALID_PARAM_GOMPERTZ': 32768,
         'INVALID_PARAM_CHISQUARE': 65536,
-        'INVALID_PARAM_DISCRETE': 131072,
-        'INVALID_PARAM_KRON_DELTA': 262144,
-        'INVALID_PARAM_DIRICHLET': 524288,
-        'INVALID_PARAM_MULTIVARIATE_STUDENT': 1048576,
-        'INVALID_PARAM_MULTINOMIAL': 2097152
+        'INVALID_PARAM_KUMARASWAMY': 131072,
+        'INVALID_PARAM_DISCRETE': 262144,
+        'INVALID_PARAM_KRON_DELTA': 524288,
+        'INVALID_PARAM_DIRICHLET': 1048576,
+        'INVALID_PARAM_MULTIVARIATE_STUDENT': 2097152,
+        'INVALID_PARAM_MULTINOMIAL': 4194304,
+        'INVALID_PARAM_BINOMIAL': 8388608,
+        'INVALID_PARAM_NEGATIVE_BINOMIAL': 16777216        
     }
     
     INVERSE_ERROR_CODES = {
@@ -370,11 +368,14 @@ class JaxRDDLCompiler:
         14: 'Found Cauchy(m, s) distribution where s <= 0.',
         15: 'Found Gompertz(k, l) distribution where either k <= 0 or l <= 0.',
         16: 'Found ChiSquare(df) distribution where df <= 0.',
-        17: 'Found Discrete(p) distribution where either p < 0 or p does not sum to 1.',
-        18: 'Found KronDelta(x) distribution where x is not int nor bool.',
-        19: 'Found Dirichlet(alpha) distribution where alpha < 0.',
-        20: 'Found MultivariateStudent(mean, cov, df) distribution where df <= 0.',
-        21: 'Found Multinomial(p, n) distribution where either p < 0, p does not sum to 1, or n <= 0.'
+        17: 'Found Kumaraswamy(a, b) distribution where either a <= 0 or b <= 0.',
+        18: 'Found Discrete(p) distribution where either p < 0 or p does not sum to 1.',
+        19: 'Found KronDelta(x) distribution where x is not int nor bool.',
+        20: 'Found Dirichlet(alpha) distribution where alpha < 0.',
+        21: 'Found MultivariateStudent(mean, cov, df) distribution where df <= 0.',
+        22: 'Found Multinomial(n, p) distribution where either p < 0, p does not sum to 1, or n <= 0.',
+        23: 'Found Binomial(n, p) distribution where either p < 0, p > 1, or n <= 0.',
+        24: 'Found NegativeBinomial(n, p) distribution where either p < 0, p > 1, or n <= 0.'        
     }
     
     @staticmethod
@@ -426,7 +427,7 @@ class JaxRDDLCompiler:
         else:
             raise RDDLNotImplementedError(
                 f'Internal error: expression type {expr} is not supported.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))
+                print_stack_trace(expr))
                 
         if dtype is not None:
             jax_expr = self._jax_cast(jax_expr, dtype)
@@ -439,9 +440,10 @@ class JaxRDDLCompiler:
         def _jax_wrapped_cast(x, key):
             val, key, err = jax_expr(x, key)
             sample = jnp.asarray(val, dtype=dtype)
-            invalid = jnp.logical_and(jnp.logical_not(jnp.can_cast(val, dtype)),
-                                      jnp.any(sample != val))
-            err |= (invalid * ERR)
+            invalid_cast = jnp.logical_and(
+                jnp.logical_not(jnp.can_cast(val, dtype)),
+                jnp.any(sample != val))
+            err |= (invalid_cast * ERR)
             return sample, key, err
         
         return _jax_wrapped_cast
@@ -455,38 +457,40 @@ class JaxRDDLCompiler:
         cached_value = self.traced.cached_sim_info(expr)
         
         def _jax_wrapped_constant(_, key):
-            return cached_value, key, NORMAL
+            sample = jnp.asarray(cached_value)
+            return sample, key, NORMAL
 
         return _jax_wrapped_constant
     
+    def _jax_pvar_slice(self, _slice):
+        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
+        
+        def _jax_wrapped_pvar_slice(x, key):
+            return _slice, key, NORMAL
+        
+        return _jax_wrapped_pvar_slice
+            
     def _jax_pvar(self, expr):
         NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
         var, pvars = expr.args  
-        cached_info = self.traced.cached_sim_info(expr)
+        is_value, cached_info = self.traced.cached_sim_info(expr)
         
         # boundary case: free variable is converted to array (0, 1, 2...)
-        if self.rddl.is_free_variable(var):
+        # boundary case: domain object is converted to canonical integer index
+        if is_value:
             cached_value = cached_info
 
-            def _jax_wrapped_pvar_free_variable(_, key):
-                return cached_value, key, NORMAL
+            def _jax_wrapped_object(_, key):
+                sample = jnp.asarray(cached_value)
+                return sample, key, NORMAL
             
-            return _jax_wrapped_pvar_free_variable
-        
-        # boundary case: enum literal is converted to canonical integer index
-        elif not pvars and self.rddl.is_literal(var):
-            cached_literal = cached_info
-            
-            def _jax_wrapped_pvar_literal(_, key):
-                return cached_literal, key, NORMAL
-            
-            return _jax_wrapped_pvar_literal
+            return _jax_wrapped_object
         
         # boundary case: no shape information (e.g. scalar pvar)
         elif cached_info is None:
             
             def _jax_wrapped_pvar_scalar(x, key):
-                sample = x[var]
+                sample = jnp.asarray(x[var])
                 return sample, key, NORMAL
             
             return _jax_wrapped_pvar_scalar
@@ -496,10 +500,11 @@ class JaxRDDLCompiler:
             slices, axis, shape, op_code, op_args = cached_info 
         
             # compile nested expressions
-            if slices and op_code == -1:
+            if slices and op_code == RDDLObjectsTracer.NUMPY_OP_CODE.NESTED_SLICE:
                 
-                jax_nested_expr = [(self._jax(arg) if _slice is None 
-                                    else (lambda _, key: (_slice, key, NORMAL)))
+                jax_nested_expr = [(self._jax(arg) 
+                                    if _slice is None 
+                                    else self._jax_pvar_slice(_slice))
                                    for (arg, _slice) in zip(pvars, slices)]    
                 
                 def _jax_wrapped_pvar_tensor_nested(x, key):
@@ -509,7 +514,8 @@ class JaxRDDLCompiler:
                     for (i, jax_expr) in enumerate(jax_nested_expr):
                         new_slices[i], key, err = jax_expr(x, key)
                         error |= err
-                    sample = sample[tuple(new_slices)]
+                    new_slices = tuple(new_slices)
+                    sample = sample[new_slices]
                     return sample, key, error
                 
                 return _jax_wrapped_pvar_tensor_nested
@@ -524,9 +530,9 @@ class JaxRDDLCompiler:
                     if axis:
                         sample = jnp.expand_dims(sample, axis=axis)
                         sample = jnp.broadcast_to(sample, shape=shape)
-                    if op_code == 0:
+                    if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.EINSUM:
                         sample = jnp.einsum(sample, *op_args)
-                    elif op_code == 1:
+                    elif op_code == RDDLObjectsTracer.NUMPY_OP_CODE.TRANSPOSE:
                         sample = jnp.transpose(sample, axes=op_args)
                     return sample, key, NORMAL
                 
@@ -546,8 +552,8 @@ class JaxRDDLCompiler:
                 sample = 1 * sample
             sample = jax_op(sample)
             if check_dtype is not None:
-                invalid = jnp.logical_not(jnp.can_cast(sample, check_dtype))
-                err |= (invalid * ERR)
+                invalid_cast = jnp.logical_not(jnp.can_cast(sample, check_dtype))
+                err |= (invalid_cast * ERR)
             return sample, key, err
         
         return _jax_wrapped_unary_op
@@ -565,10 +571,10 @@ class JaxRDDLCompiler:
             sample = jax_op(sample1, sample2)
             err = err1 | err2
             if check_dtype is not None:
-                invalid = jnp.logical_not(jnp.logical_and(
+                invalid_cast = jnp.logical_not(jnp.logical_and(
                     jnp.can_cast(sample1, check_dtype),
                     jnp.can_cast(sample2, check_dtype)))
-                err |= (invalid * ERR)
+                err |= (invalid_cast * ERR)
             return sample, key, err
         
         return _jax_wrapped_binary_op
@@ -653,8 +659,8 @@ class JaxRDDLCompiler:
             if is_floating:
                 sample = 1 * sample
             else:
-                invalid = jnp.logical_not(jnp.can_cast(sample, bool))
-                err |= (invalid * ERR)
+                invalid_cast = jnp.logical_not(jnp.can_cast(sample, bool))
+                err |= (invalid_cast * ERR)
             sample = jax_op(sample, axis=axes)
             return sample, key, err
         
@@ -683,27 +689,30 @@ class JaxRDDLCompiler:
                 jax_lhs, jax_rhs, jax_op, at_least_int=True)
         
         raise RDDLNotImplementedError(
-                f'Function {op} is not supported.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))   
+                f'Function {op} is not supported.\n' + print_stack_trace(expr))   
     
     # ===========================================================================
     # control flow
     # ===========================================================================
     
     def _jax_control(self, expr):
-        _, op = expr.etype
-        valid_ops = self.CONTROL_OPS
-        JaxRDDLCompiler._check_valid_op(expr, valid_ops)
-        
+        _, op = expr.etype        
         if op == 'if':
             return self._jax_if(expr)
-        else:
+        elif op == 'switch':
             return self._jax_switch(expr)
+        
+        raise RDDLNotImplementedError(
+                f'Control operator {op} is not supported.\n' + 
+                print_stack_trace(expr))   
+    
+    def _jax_if_helper(self):
+        return jnp.where
     
     def _jax_if(self, expr):
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_CAST']
         JaxRDDLCompiler._check_num_args(expr, 3)
-        jax_op = self.CONTROL_OPS['if']
+        jax_if = self._jax_if_helper()
         
         pred, if_true, if_false = expr.args        
         jax_pred = self._jax(pred)
@@ -714,10 +723,10 @@ class JaxRDDLCompiler:
             sample1, key, err1 = jax_pred(x, key)
             sample2, key, err2 = jax_true(x, key)
             sample3, key, err3 = jax_false(x, key)
-            sample = jax_op(sample1, sample2, sample3)
+            sample = jax_if(sample1, sample2, sample3)
             err = err1 | err2 | err3
-            invalid = jnp.logical_not(jnp.can_cast(sample1, bool))
-            err |= (invalid * ERR)
+            invalid_cast = jnp.logical_not(jnp.can_cast(sample1, bool))
+            err |= (invalid_cast * ERR)
             return sample, key, err
             
         return _jax_wrapped_if_then_else
@@ -735,7 +744,7 @@ class JaxRDDLCompiler:
     def _jax_switch(self, expr):
         pred, *_ = expr.args             
         jax_pred = self._jax(pred)
-        jax_switch = self.CONTROL_OPS['switch']
+        jax_switch = self._jax_switch_helper()
         
         # wrap cases as JAX expressions
         cases, default = self.traced.cached_sim_info(expr) 
@@ -765,6 +774,34 @@ class JaxRDDLCompiler:
     # random variables
     # ===========================================================================
     
+    # distributions with complete reparameterization support:
+    # KronDelta: complete
+    # DiracDelta: complete
+    # Uniform: complete
+    # Bernoulli: complete (subclass uses Gumbel-softmax)
+    # Normal: complete
+    # Exponential: complete
+    # Weibull: complete
+    # Pareto: complete
+    # Gumbel: complete
+    # Laplace: complete
+    # Cauchy: complete
+    # Gompertz: complete
+    # Kumaraswamy: complete
+    # Discrete: complete (subclass uses Gumbel-softmax)
+    # UnnormDiscrete: complete (subclass uses Gumbel-softmax)
+    # Discrete(p): complete (subclass uses Gumbel-softmax)
+    # UnnormDiscrete(p): complete (subclass uses Gumbel-softmax)
+    
+    # distributions with incomplete reparameterization support (TODO):
+    # Binomial: (use truncation and Gumbel-softmax)
+    # NegativeBinomial: (no reparameterization)
+    # Poisson: (use truncation and Gumbel-softmax)
+    # Gamma, ChiSquare: (no shape reparameterization)
+    # Beta: (no reparameterization)
+    # Geometric: (implement safe floor)
+    # Student: (no reparameterization)
+    
     def _jax_random(self, expr):
         _, name = expr.etype
         if name == 'KronDelta':
@@ -773,18 +810,22 @@ class JaxRDDLCompiler:
             return self._jax_dirac(expr)
         elif name == 'Uniform':
             return self._jax_uniform(expr)
+        elif name == 'Bernoulli':
+            return self._jax_bernoulli(expr)
         elif name == 'Normal':
-            return self._jax_normal(expr)
+            return self._jax_normal(expr)  
+        elif name == 'Poisson':
+            return self._jax_poisson(expr)
         elif name == 'Exponential':
             return self._jax_exponential(expr)
         elif name == 'Weibull':
-            return self._jax_weibull(expr)   
-        elif name == 'Bernoulli':
-            return self._jax_bernoulli(expr)
-        elif name == 'Poisson':
-            return self._jax_poisson(expr)
+            return self._jax_weibull(expr) 
         elif name == 'Gamma':
             return self._jax_gamma(expr)
+        elif name == 'Binomial':
+            return self._jax_binomial(expr)
+        elif name == 'NegativeBinomial':
+            return self._jax_negative_binomial(expr)
         elif name == 'Beta':
             return self._jax_beta(expr)
         elif name == 'Geometric':
@@ -803,6 +844,8 @@ class JaxRDDLCompiler:
             return self._jax_gompertz(expr)
         elif name == 'ChiSquare':
             return self._jax_chisquare(expr)
+        elif name == 'Kumaraswamy':
+            return self._jax_kumaraswamy(expr)
         elif name == 'Discrete':
             return self._jax_discrete(expr, unnorm=False)
         elif name == 'UnnormDiscrete':
@@ -814,7 +857,7 @@ class JaxRDDLCompiler:
         else:
             raise RDDLNotImplementedError(
                 f'Distribution {name} is not supported.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))
+                print_stack_trace(expr))
         
     def _jax_kron(self, expr):
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_KRON_DELTA']
@@ -825,8 +868,9 @@ class JaxRDDLCompiler:
         # just check that the sample can be cast to int
         def _jax_wrapped_distribution_kron(x, key):
             sample, key, err = arg(x, key)
-            invalid = jnp.logical_not(jnp.can_cast(sample, JaxRDDLCompiler.INT))
-            err |= (invalid * ERR)
+            invalid_cast = jnp.logical_not(
+                jnp.can_cast(sample, JaxRDDLCompiler.INT))
+            err |= (invalid_cast * ERR)
             return sample, key, err
                         
         return _jax_wrapped_distribution_kron
@@ -923,10 +967,14 @@ class JaxRDDLCompiler:
             return sample, key, err
         
         return _jax_wrapped_distribution_weibull
-            
+    
+    def _jax_bernoulli_helper(self):
+        return random.bernoulli
+        
     def _jax_bernoulli(self, expr):
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_BERNOULLI']
         JaxRDDLCompiler._check_num_args(expr, 1)
+        bernoulli = self._jax_bernoulli_helper()
         
         arg_prob, = expr.args
         jax_prob = self._jax(arg_prob)
@@ -935,7 +983,7 @@ class JaxRDDLCompiler:
         def _jax_wrapped_distribution_bernoulli(x, key):
             prob, key, err = jax_prob(x, key)
             key, subkey = random.split(key)
-            sample = random.bernoulli(key=subkey, p=prob)
+            sample = bernoulli(subkey, prob)
             out_of_bounds = jnp.logical_not(jnp.all((prob >= 0) & (prob <= 1)))
             err |= (out_of_bounds * ERR)
             return sample, key, err
@@ -983,6 +1031,55 @@ class JaxRDDLCompiler:
         
         return _jax_wrapped_distribution_gamma
     
+    def _jax_binomial(self, expr):
+        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_BINOMIAL']
+        JaxRDDLCompiler._check_num_args(expr, 2)
+        
+        arg_trials, arg_prob = expr.args
+        jax_trials = self._jax(arg_trials)
+        jax_prob = self._jax(arg_prob)
+        
+        # uses the JAX substrate of tensorflow-probability
+        def _jax_wrapped_distribution_binomial(x, key):
+            trials, key, err2 = jax_trials(x, key)       
+            prob, key, err1 = jax_prob(x, key)
+            trials = jnp.asarray(trials, JaxRDDLCompiler.REAL)
+            prob = jnp.asarray(prob, JaxRDDLCompiler.REAL)
+            key, subkey = random.split(key)
+            dist = tfp.distributions.Binomial(total_count=trials, probs=prob)
+            sample = dist.sample(seed=subkey).astype(JaxRDDLCompiler.INT)
+            out_of_bounds = jnp.logical_not(jnp.all(
+                (prob >= 0) & (prob <= 1) & (trials >= 0)))
+            err = err1 | err2 | (out_of_bounds * ERR)
+            return sample, key, err
+        
+        return _jax_wrapped_distribution_binomial
+    
+    def _jax_negative_binomial(self, expr):
+        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_NEGATIVE_BINOMIAL']
+        JaxRDDLCompiler._check_num_args(expr, 2)
+        
+        arg_trials, arg_prob = expr.args
+        jax_trials = self._jax(arg_trials)
+        jax_prob = self._jax(arg_prob)
+        
+        # uses the JAX substrate of tensorflow-probability
+        def _jax_wrapped_distribution_negative_binomial(x, key):
+            trials, key, err2 = jax_trials(x, key)       
+            prob, key, err1 = jax_prob(x, key)
+            trials = jnp.asarray(trials, JaxRDDLCompiler.REAL)
+            prob = jnp.asarray(prob, JaxRDDLCompiler.REAL)
+            key, subkey = random.split(key)
+            dist = tfp.distributions.NegativeBinomial(
+                total_count=trials, probs=prob)
+            sample = dist.sample(seed=subkey).astype(JaxRDDLCompiler.INT)
+            out_of_bounds = jnp.logical_not(jnp.all(
+                (prob >= 0) & (prob <= 1) & (trials > 0)))
+            err = err1 | err2 | (out_of_bounds * ERR)
+            return sample, key, err
+        
+        return _jax_wrapped_distribution_negative_binomial    
+        
     def _jax_beta(self, expr):
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_BETA']
         JaxRDDLCompiler._check_num_args(expr, 2)
@@ -1170,18 +1267,40 @@ class JaxRDDLCompiler:
         
         return _jax_wrapped_distribution_chisquare
     
+    def _jax_kumaraswamy(self, expr):
+        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_KUMARASWAMY']
+        JaxRDDLCompiler._check_num_args(expr, 2)
+        
+        arg_a, arg_b = expr.args
+        jax_a = self._jax(arg_a)
+        jax_b = self._jax(arg_b)
+        
+        # uses the reparameterization K(a, b) = (1 - (1 - U(0, 1))^{1/b})^{1/a}
+        def _jax_wrapped_distribution_kumaraswamy(x, key):
+            a, key, err1 = jax_a(x, key)
+            b, key, err2 = jax_b(x, key)
+            key, subkey = random.split(key)
+            U = random.uniform(
+                key=subkey, shape=jnp.shape(a), dtype=JaxRDDLCompiler.REAL)            
+            sample = jnp.power(1.0 - jnp.power(U, 1.0 / b), 1.0 / a)
+            out_of_bounds = jnp.logical_not(jnp.all((a > 0) & (b > 0)))
+            err = err1 | err2 | (out_of_bounds * ERR)
+            return sample, key, err
+        
+        return _jax_wrapped_distribution_kumaraswamy
+    
     # ===========================================================================
     # random variables with enum support
     # ===========================================================================
     
     def _jax_discrete_helper(self):
         
-        def _jax_discrete_calc_exact(prob, subkey):
+        def _jax_discrete_calc_exact(key, prob):
+            logits = jnp.log(prob)
+            sample = random.categorical(key=key, logits=logits, axis=-1)
             out_of_bounds = jnp.logical_not(jnp.logical_and(
                 jnp.all(prob >= 0),
                 jnp.allclose(jnp.sum(prob, axis=-1), 1.0)))
-            logits = jnp.log(prob)
-            sample = random.categorical(key=subkey, logits=logits, axis=-1)
             return sample, out_of_bounds
         
         return _jax_discrete_calc_exact
@@ -1189,10 +1308,10 @@ class JaxRDDLCompiler:
     def _jax_discrete(self, expr, unnorm):
         NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_DISCRETE']
+        jax_discrete = self._jax_discrete_helper()
         
         ordered_args = self.traced.cached_sim_info(expr)
         jax_probs = [self._jax(arg) for arg in ordered_args]
-        jax_discrete = self._jax_discrete_helper()
         
         def _jax_wrapped_distribution_discrete(x, key):
             
@@ -1209,7 +1328,7 @@ class JaxRDDLCompiler:
             
             # dispatch to sampling subroutine
             key, subkey = random.split(key)
-            sample, out_of_bounds = jax_discrete(prob, subkey)
+            sample, out_of_bounds = jax_discrete(subkey, prob)
             error |= (out_of_bounds * ERR)
             return sample, key, error
         
@@ -1218,11 +1337,11 @@ class JaxRDDLCompiler:
     def _jax_discrete_pvar(self, expr, unnorm):
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_DISCRETE']
         JaxRDDLCompiler._check_num_args(expr, 1)
+        jax_discrete = self._jax_discrete_helper()
         
         _, args = expr.args
         arg, = args
         jax_probs = self._jax(arg)
-        jax_discrete = self._jax_discrete_helper()
 
         def _jax_wrapped_distribution_discrete_pvar(x, key):
             
@@ -1234,7 +1353,7 @@ class JaxRDDLCompiler:
             
             # dispatch to sampling subroutine
             key, subkey = random.split(key)
-            sample, out_of_bounds = jax_discrete(prob, subkey)
+            sample, out_of_bounds = jax_discrete(subkey, prob)
             error |= (out_of_bounds * ERR)
             return sample, key, error
         
@@ -1257,7 +1376,7 @@ class JaxRDDLCompiler:
         else:
             raise RDDLNotImplementedError(
                 f'Distribution {name} is not supported.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))
+                print_stack_trace(expr))
     
     def _jax_multivariate_normal(self, expr): 
         _, args = expr.args
@@ -1346,53 +1465,29 @@ class JaxRDDLCompiler:
         return _jax_wrapped_distribution_dirichlet
     
     def _jax_multinomial(self, expr):
-        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
         ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_MULTINOMIAL']
         
         _, args = expr.args
-        prob, trials = args
-        jax_prob = self._jax(prob)
+        trials, prob = args
         jax_trials = self._jax(trials)
-        jax_discrete = self._jax_discrete_helper()
+        jax_prob = self._jax(prob)
         index, = self.traced.cached_sim_info(expr)
         
-        # samples from a discrete(prob), computes a one-hot mask w.r.t categories
-        def _jax_wrapped_multinomial_trial(_, values):
-            samples, errors, key, prob, categories = values
-            key, subkey = random.split(key)
-            sample, error = jax_discrete(prob, subkey)            
-            sample = sample[jnp.newaxis, ...]
-            masked = (sample == categories)
-            samples += masked
-            errors |= error
-            return (samples, errors, key, prob, categories)            
-            
         def _jax_wrapped_distribution_multinomial(x, key):
-            
-            # sample multinomial parameters
-            # note that # of trials must not be parameterized
-            prob, key, err1 = jax_prob(x, key)
-            trials, key, err2 = jax_trials(x, key)
-            num_trials = jnp.ravel(trials)[0]
-            out_of_bounds = jnp.logical_not(num_trials > 0)
+            trials, key, err1 = jax_trials(x, key)
+            prob, key, err2 = jax_prob(x, key)
+            trials = jnp.asarray(trials, JaxRDDLCompiler.REAL)
+            prob = jnp.asarray(prob, JaxRDDLCompiler.REAL)
+            key, subkey = random.split(key)
+            dist = tfp.distributions.Multinomial(total_count=trials, probs=prob)
+            sample = dist.sample(seed=subkey).astype(JaxRDDLCompiler.INT)
+            sample = jnp.moveaxis(sample, source=-1, destination=index)
+            out_of_bounds = jnp.logical_not(jnp.all(
+                (prob >= 0) & 
+                jnp.allclose(jnp.sum(prob, axis=-1), 1.0) & 
+                (trials >= 0)))
             error = err1 | err2 | (out_of_bounds * ERR)
-            
-            # create a categories mask that spans support of discrete(prob)
-            num_categories = prob.shape[-1]
-            categories = np.arange(num_categories)
-            categories = categories[(...,) + (jnp.newaxis,) * len(prob.shape[:-1])]
-            
-            # do a for loop over trials, accumulate category counts in samples
-            counts = jnp.zeros(shape=(num_categories,) + prob.shape[:-1],
-                               dtype=JaxRDDLCompiler.INT)
-            sample, err3, key, *_ = jax.lax.fori_loop(
-                lower=0,
-                upper=num_trials,
-                body_fun=_jax_wrapped_multinomial_trial,
-                init_val=(counts, NORMAL, key, prob, categories))
-            sample = jnp.moveaxis(sample, source=0, destination=index)
-            error |= err3
-            return sample, key, error
+            return sample, key, error            
         
         return _jax_wrapped_distribution_multinomial
     
@@ -1405,11 +1500,13 @@ class JaxRDDLCompiler:
         if op == 'det':
             return self._jax_matrix_det(expr)
         elif op == 'inverse':
-            return self._jax_matrix_inv(expr)
+            return self._jax_matrix_inv(expr, pseudo=False)
+        elif op == 'pinverse':
+            return self._jax_matrix_inv(expr, pseudo=True)
         else:
             raise RDDLNotImplementedError(
                 f'Matrix operation {op} is not supported.\n' + 
-                JaxRDDLCompiler._print_stack_trace(expr))
+                print_stack_trace(expr))
     
     def _jax_matrix_det(self, expr):
         * _, arg = expr.args
@@ -1422,14 +1519,15 @@ class JaxRDDLCompiler:
         
         return _jax_wrapped_matrix_operation_det
     
-    def _jax_matrix_inv(self, expr):
+    def _jax_matrix_inv(self, expr, pseudo):
         _, arg = expr.args
         jax_arg = self._jax(arg)
         indices = self.traced.cached_sim_info(expr)
+        op = jnp.linalg.pinv if pseudo else jnp.linalg.inv
         
         def _jax_wrapped_matrix_operation_inv(x, key):
             sample_arg, key, error = jax_arg(x, key)
-            sample = jnp.linalg.inv(sample_arg)
+            sample = op(sample_arg)
             sample = jnp.moveaxis(sample, source=(-2, -1), destination=indices)
             return sample, key, error
         
