@@ -3,6 +3,15 @@ import os
 from PIL import Image
 import warnings
 
+# (mike: #166) opencv is now optional
+try:
+    import cv2
+    _ALLOW_MP4 = True
+except:
+    warnings.warn('cv2 is not installed: save_as_mp4 option will be disabled.',
+                  stacklevel=2)
+    _ALLOW_MP4 = False
+
 
 class MovieGenerator:
     
@@ -11,18 +20,20 @@ class MovieGenerator:
                  env_name: str,
                  max_frames: int,
                  skip: int=1,
-                 frame_duration: int=50,
-                 loop: int=0, 
-                 save_format: str='png'):
+                 save_format: str='png',
+                 frame_duration: int=100,
+                 loop: int=0,
+                 save_as_mp4: bool=False):
         '''Creates a new movie generator for saving frames to disk, and creating animated GIFs.
         
         :param save_dir: the directory to save images to
         :param env_name: the root name of each image file
         :param max_frames: the max number of frames to save
         :param skip: how often frames should be recorded
-        :param frame_duration: the duration of each frame in the animated GIF
-        :param loop: how many times the animated GIF should loop
         :param save_format: the format in which to save individual frames
+        :param frame_duration: the duration of each frame in the animated video
+        :param loop: how many times the animated GIF should loop
+        :param save_as_mp4: whether to save mp4 video (or GIF if False)
         '''
         self.save_dir = save_dir
         self.save_path = os.path.join(save_dir, env_name + '_{}' + '.' + save_format)
@@ -31,6 +42,7 @@ class MovieGenerator:
         self.skip = skip
         self.frame_duration = frame_duration
         self.loop = loop
+        self.save_as_mp4 = save_as_mp4
         
         self._n_frame = 0
         self._time = 0
@@ -43,8 +55,8 @@ class MovieGenerator:
             os.remove(file)
             removed += 1
         if removed:
-            warnings.warn(f'removed {removed} temporary files at {load_path}', 
-                          FutureWarning, stacklevel=2)
+            warnings.warn(f'removed {removed} temporary files at {load_path}',
+                          stacklevel=2)
         
         self._n_frame = 0
         self._time = 0
@@ -62,6 +74,13 @@ class MovieGenerator:
         self._n_frame += 1 
         self._time += 1
     
+    def save_animation(self, file_name: str=None):
+        if _ALLOW_MP4 and self.save_as_mp4:
+            self.save_mp4(file_name)
+        else:
+            self.save_gif(file_name)
+        self.reset()
+            
     def save_gif(self, file_name: str=None):
         if file_name is None:
             file_name = self.env_name
@@ -76,6 +95,22 @@ class MovieGenerator:
                         append_images=images,
                         save_all=True,
                         duration=self.frame_duration,
-                        loop=self.loop)
+                        loop=self.loop)  
         
-        self.reset()
+    def save_mp4(self, file_name: str=None):
+        if file_name is None:
+            file_name = self.env_name
+        load_path = self.save_path.format('*')
+        images = glob.glob(load_path)
+        
+        writer, w, h = None, None, None
+        fps = 1000 / self.frame_duration
+        for file in images:
+            frame = cv2.imread(file)
+            if w is None:
+                h, w, _ = frame.shape
+                fourcc = cv2.VideoWriter_fourcc('m', 'p', '4', 'v')
+                writer = cv2.VideoWriter(file_name + '.mp4', fourcc, fps, (w, h))
+            writer.write(frame)
+        if writer is not None:
+            writer.release()
